testECCCR <- function () {

  numInstances = as.integer(500)
  numAttribs = as.integer(100)
  numLabels = as.integer(5)

  repetitions = as.integer(3)

  buildSize = as.integer(numInstances * 0.67)
  testSize = as.integer(numInstances * 0.33)
  k = as.integer(10)

  valArray = sample(-5:5, numInstances * numAttribs, replace=T)
  labArray = sample(0:1, numInstances * numLabels, replace=T)

  dim(valArray) <- c(numInstances, numAttribs)
  dim(labArray) <- c(numInstances, numLabels)

  values = as.data.frame(valArray)
  labels = as.data.frame(labArray)

  resultFrame = eccCLcrossValidate(values, labels, numInstances,
                     repetitions,
                     buildSize, testSize, k
                     )

  print(resultFrame)

}

testECCCRbyFile <- function (file) {

  repetitions = as.integer(3)

  buildSize = as.integer(400)
  testSize = as.integer(100)
  k = as.integer(10)

  wekaLoader = eccCLloadWekaFile(file)
  values = eccCLwekaValues(wekaLoader)
  labels = eccCLwekaLabels(wekaLoader)
  numInstances = eccCLwekaNumInstances(wekaLoader)

  resultFrame = eccCLcrossValidate(values, labels, numInstances,
                                   repetitions,
                                   buildSize, testSize, k
                                   )

  print(resultFrame)

}

testBasic <- function(trainFile, testFile) {
  wekaBuild <- eccCLloadWekaFile(trainFile)
  buildValues <- eccCLwekaValues(wekaBuild)
  buildLabels <- eccCLwekaLabels(wekaBuild)
  buildInstances <- eccCLwekaNumInstances(wekaBuild)

  ecc <- eccCLbuild(buildValues, buildLabels, buildInstances)

  wekaTest <- eccCLloadWekaFile(testFile)
  testValues <- eccCLwekaValues(wekaTest)
  testLabels <- eccCLwekaLabels(wekaTest)
  testInstances <- eccCLwekaNumInstances(wekaTest)

  resultsRaw <- eccCLclassify(ecc, testValues, testLabels, testInstances)
  results <- eccCLgetResults(resultsRaw)

  print(results)
}

testBuildAndStore <- function(dataFile, eccFile) {
  wekaBuild <- eccCLloadWekaFile(dataFile)
  buildValues <- eccCLwekaValues(wekaBuild)
  buildLabels <- eccCLwekaLabels(wekaBuild)
  buildInstances <- eccCLwekaNumInstances(wekaBuild)

  ecc <- eccCLbuild(buildValues, buildLabels, buildInstances)
  eccCLstore(ecc, eccFile)
}

testLoadAndClassify <- function(dataFile, eccFile) {
  wekaTest <- eccCLloadWekaFile(dataFile)
  testValues <- eccCLwekaValues(wekaTest)
  testLabels <- eccCLwekaLabels(wekaTest)
  testInstances <- eccCLwekaNumInstances(wekaTest)

  ecc <- eccCLload(eccFile)

  resultsRaw <- eccCLclassify(ecc, testValues, testLabels, testInstances)
  results <- eccCLgetResults(resultsRaw)

  print(results)
}
